#pragma once

#include <mutex>

#include <QObject>

#include "RobotAPI/libraries/skills/core/ProviderID.h"
#include "RobotAPI/libraries/skills/core/SkillDescription.h"
#include "RobotAPI/libraries/skills/core/SkillExecutionID.h"
#include "RobotAPI/libraries/skills/core/SkillStatusUpdate.h"
#include <RobotAPI/interface/skills/SkillManagerInterface.h>

namespace armarx::skills::gui
{
    using StatusMap = std::map<skills::SkillExecutionID, skills::SkillStatusUpdate>;
    using SkillMap =
        std::map<skills::ProviderID, std::map<skills::SkillID, skills::SkillDescription>>;

    class SkillManagerWrapper : public QObject
    {
        Q_OBJECT
    public:
        SkillManagerWrapper(skills::manager::dti::SkillManagerInterfacePrx& _memory,
                            QObject* parent = nullptr) :
            QObject(parent), memory(_memory)
        {
        }

        // We can instantiate the proxy without memory, if it's not connected yet.
        SkillManagerWrapper(QObject* parent = nullptr) : QObject(parent)
        {
        }

        /**
         * @brief Updates the memory pointer. This should be called whenever the GUI connects/reconnects.
         */
        void connectMemory(skills::manager::dti::SkillManagerInterfacePrx const& updatedMemory);

        // Basic functions to work with. These will produce Ice calls!
        /**
         * @brief Attempts to start an execution with given parameters. If an Ice exception is encountered, a message will be logged and the method will abort.
         * @param skillId The skill to execute.
         * @param params The parameters to execute with.
         */
        void startExecutionWithParams(skills::SkillID& skillId, aron::data::DictPtr const params);

        /**
         * @brief Attempts to stop an execution. If this doesn't succeed, it will retry.
         * @param executionId The ExecutionID to abort.
         * @param max_retries The amount of times to retry, if aborting fails.
         */
        void stopExecution(skills::SkillExecutionID const& executionId,
                           const unsigned int max_retries = 0);

        /**
         * @brief Fetches and returns the latest skills update from memory.
         * @return The map representing all skills in memory. Empty, if error occurred.
         */
        const SkillMap fetchSkills();

        /**
         * @brief Fetches and returns the latest status update from memory.
         * @return The map containing status updates for all execution ids. Empty, if error occurred.
         */
        const StatusMap fetchExecutions();

        static const std::optional<skills::ProviderID> findFirstProvider(SkillMap const& map,
                                                                         SkillID const& skillId);

    signals:
        void connectionUpdate(std::string const& message, std::string const& error);

    public slots:
        /**
         * @brief Disconnects the interface from memory.
         */
        void disconnectMemory();

        /**
         * @brief Applies the search word to the update filter.
         * @param search The search word.
         */
        void acceptSearchRequest(std::string const& search);

        /**
         * @brief Stops all available (and running) executions.
         */
        void stopAllExecutions();

    private:
        mutable std::mutex mutex_memory;

        armarx::skills::manager::dti::SkillManagerInterfacePrx memory;
        std::string currentSkillSearch = "";

        /**
         * @brief Modifies the given map to match the search.
         * @param update The map to modify.
         */
        void filterSkillUpdate(std::map<skills::manager::dto::SkillID,
                                        skills::manager::dto::SkillDescription>& update);
    };
} // namespace armarx::skills::gui
